#include <string>
#include <iostream>
#include <fstream>

#include "SmartAlgorithm.h"
#include "Choice.h"
#include "Engine.h"

#define DEBUG
//#undef DEBUG

using namespace std;

SmartAlgorithm::SmartAlgorithm()
    : patternLength(PATTERN_LENGTH), history(""),
      filename(FREQUENCY_FILE_NAME),
      random_algorithm(new RandomAlgorithm())
{
    read_frequencies();
}

SmartAlgorithm::~SmartAlgorithm()
{
    write_frequencies();

    delete random_algorithm;
}

Choice SmartAlgorithm::predict()
{
    int length = history.length();
    int max_freq = 0;

    round = engine->get_round();

    // Use the recorded frequencies to predict the human's choice.
    if (length >= patternLength)
    {
        string most_recent = history.substr(1);
        string r_pattern = most_recent + 'R';
        string p_pattern = most_recent + 'P';
        string s_pattern = most_recent + 'S';

        int r_freq = check_frequency(r_pattern);
        int p_freq = check_frequency(p_pattern);
        int s_freq = check_frequency(s_pattern);

        // Choose the pattern with the highest frequency.
        if (r_freq > max_freq)
        {
            max_freq = r_freq;
            prediction = Choice::ROCK;
        }

        if (p_freq > max_freq)
        {
            max_freq = p_freq;
            prediction = Choice::PAPER;
        }

        if (s_freq > max_freq)
        {
            max_freq = s_freq;
            prediction = Choice::SCISSORS;
        }

        if (max_freq == 0) prediction = Choice::NONE;

#ifdef DEBUG
        cout << "    " << round << "-Computer predicts "
             << to_string(prediction) << endl;
#endif

        return prediction;
    }

    // Insufficient history to predict.
    else
    {
        return Choice::NONE;
    }
}

Choice SmartAlgorithm::decide(const Choice prediction)
{
    Choice computerChoice;

    // Make the choice based on a prediction if possible,
    // else just make a random choice.
    switch (prediction)
    {
        case Choice::ROCK:     computerChoice = Choice::PAPER;    break;
        case Choice::PAPER:    computerChoice = Choice::SCISSORS; break;
        case Choice::SCISSORS: computerChoice = Choice::ROCK;     break;

        default: computerChoice = random_algorithm->choose();
    }

    Choice humanChoice = engine->get_human_choice();

#ifdef DEBUG
    cout << "    " << round << "-Computer chooses " << computerChoice << endl;
    cout << "    " << round << "-Human chose " << humanChoice << endl;
#endif

    update_frequency(humanChoice, computerChoice);
    return computerChoice;
}

int SmartAlgorithm::check_frequency(const string history) const
{
    auto it = frequencies.find(history);
    int freq = it != frequencies.end() ? it->second : 0;

#ifdef DEBUG
    cout << "        Checked: " << history << ":" << freq << endl;
#endif

    return freq;
}

void SmartAlgorithm::update_frequency(const Choice humanChoice,
                                      const Choice computerChoice)
{
    append_history(Who::HUMAN, humanChoice);

    int length = history.length();
    if (length >= patternLength)
    {
        auto it = frequencies.find(history);

        if (it == frequencies.end())
        {
            frequencies.insert(pair<string, int>(history, 1));

#ifdef DEBUG
            cout << "        Frequency inserted " << history << ":1 size="
                 << frequencies.size() << endl;
#endif
        }
        else
        {
            it->second++;

#ifdef DEBUG
            cout << "        Frequency updated " << history << ":"
                 << it->second << endl;
#endif
        }
    }

    append_history(Who::COMPUTER, computerChoice);
}

void SmartAlgorithm::append_history(const Who who, const Choice choice)
{
    if (choice != Choice::NONE)
    {
#ifdef DEBUG
        cout << "        History appended by " << to_string(who)
             << " " << to_char(choice) << endl;
#endif

        history.push_back(to_char(choice));
        if (history.length() > patternLength) history.erase(0, 1);

#ifdef DEBUG
        cout << "        History now " << history << endl;
#endif
    }
}

void SmartAlgorithm::read_frequencies()
{
    ifstream freqIn;
    freqIn.open(filename);

    if (freqIn.fail())
    {
        cout << "No existing frequency file " << filename << "." << endl;
        history = "";
    }
    else
    {
#ifdef DEBUG
        cout << "Reading frequency file " << filename << ":" << endl;
#endif

        freqIn >> history;

#ifdef DEBUG
        cout << "  History: " << history << endl;
        int n = 0;
#endif

        string str;
        int count;
        do
        {
            freqIn >> str >> count;
            if (!freqIn.eof())
            {
                frequencies[str] = count;

#ifdef DEBUG
                cout << "  " << str << " " << count << endl;
                n++;
#endif

            }
        } while (!freqIn.eof());

#ifdef DEBUG
        cout << "Read " << n << " patterns." << endl;
#endif

        freqIn.close();
    }
}

void SmartAlgorithm::write_frequencies()
{
    ofstream freqOut;
    freqOut.open(filename);

    if (freqOut.fail())
    {
        cout << "Cannot write file " << filename << endl;
    }
    else
    {
#ifdef DEBUG
        cout << "Writing frequency file "<< filename << ":" << endl;
#endif

        freqOut << history << endl;

#ifdef DEBUG
        cout << "  History: " << history << endl;
        int n = 0;
#endif

        for (auto it = frequencies.begin(); it != frequencies.end(); it++)
        {
            freqOut << it->first << " " << it->second << endl;

#ifdef DEBUG
            cout << "  " << it->first << " " << it->second << endl;
            n++;
#endif

        }

#ifdef DEBUG
        cout << "Wrote " << n << " patterns." << endl;
#endif

        freqOut.close();
    }

    cout << endl;
}
